#!/usr/bin/env python
# coding=utf-8
#
# Copyright © Splunk, Inc. All Rights Reserved.

from __future__ import absolute_import, division, print_function, unicode_literals

from collections import OrderedDict
from io import StringIO
from os import access, chmod, environ, mkdir, path
from tempfile import mkdtemp

from slim.app import AppConfiguration
from slim.utils import SlimLogger
from slim.utils.internal import string

from shutil import rmtree
from unittest import TestCase

from test_slim import download_archive

import io
import os
import json
import stat
import tarfile

# TODO: [ ] Setting declaration pattern compiler SHOULD create an '<__any_name>' pattern from any pattern that looks
# like this:
# {code}
# <name> = <value>
# {code}
# Requirement: the '<__any_name>' should compile to r'(?P<__any_name>.+?)\Z. Note the '+' which requires one or more
# matching characters.
# Rationale:
# * Enforces the requirement that you must provide a setting name (i.e., you can't say this: ' = some value'
# * Merges <name> = <value> patterns
# We say SHOULD, not MUST because
# * This change is not required because the conf parser enforces the requirement that you must provide a setting name
# * The benefit of merging <name> = <value> patterns in the conf spec processor is negligible.

# TODO: [ ] app._internal.object_view.ObjectView should either not be used by command.public or ObjectView should be
# moved to another module; perhaps utils.object_view or collections.object_view where we might also have
# collections.ordered_set


class TestAppConfiguration(TestCase):

    # region Tests

    def test_load(self):

        handler = SlimLogger.handlers()[0]
        current_stream = handler.stream
        handler.stream = StringIO()

        try:
            app_configuration = AppConfiguration.load(self._test_app_root)
            observed_messages = handler.stream.getvalue()
        finally:
            handler.stream = current_stream

        # Compare logger messages

        app_root = self._test_app_root

        with io.open(path.join(app_root, '..', 'expected_messages.txt'), encoding='utf-8') as istream:
            expected_messages = istream.read() % {'app_root_local': path.join(app_root, 'local', '')}

        self.assertTextEqual(expected_messages, observed_messages, 'messages')

        # Compare Configuration JSON files

        observed_configuration_json_filename = path.join(self._output_dir, 'observed_configuration.json')
        app_configuration.save(observed_configuration_json_filename)

        with io.open(observed_configuration_json_filename, encoding='utf-8') as istream:
            observed_configuration_json = istream.read()

        expected_configuration_json_filename = path.join(app_root, '..', 'expected_configuration.json')

        with io.open(expected_configuration_json_filename, encoding='utf-8') as istream:
            expected_configuration_json = istream.read() % {
                'app_root_default': path.join(app_root, 'default', '').replace('\\', '\\\\'),
                'app_root_local': path.join(app_root, 'local', '').replace('\\', '\\\\')
            }

        self.assertJsonEqual(expected_configuration_json, observed_configuration_json, 'configuration')
        return

    # endregion

    # region Setup and teardown

    @staticmethod
    def onerror(function, filename, exception_info):
        if access(filename, os.W_OK):
            raise exception_info[1]
        chmod(filename, stat.S_IWUSR)
        function(filename)

    @classmethod
    def setUpClass(cls):
        cls._output_dir = mkdtemp(prefix='slim-' + cls._script_name + '-', dir=environ['SPLUNK_SOURCE'])
        if path.isdir(cls._test_app_root):
            rmtree(cls._test_app_root, onerror=cls.onerror)
        # Download the source package before we extract it
        download_file = '@' + os.path.join(cls._test_dir, 'download.args')
        download_archive(test_dir=cls._script_root, download_file=download_file)
        with tarfile.open(cls._test_app_root + '.tar.gz') as archive:
            archive.extractall(path=cls._test_dir)
        return

    @classmethod
    def tearDownClass(cls):
        rmtree(cls._test_app_root, onerror=cls.onerror)
        rmtree(cls._output_dir, onerror=cls.onerror)
        os.remove(cls._test_app_root + '.tar.gz')

    # endregion

    # region Assertions

    def assertJsonEqual(self, expected_result, observed_result, base_name):

        if type(observed_result) is string:
            observed_result = json.loads(observed_result, object_pairs_hook=self._to_ordered_dict)

        if type(expected_result) is string:
            expected_result = json.loads(expected_result, object_pairs_hook=self._to_ordered_dict)

        directory_name = path.join(self._test_dir, 'observed-results')
        expected_result_filename = base_name + '-expected.json'
        observed_result_filename = base_name + '-observed.json'

        error_message = base_name + ': observed value is different than expected:\n' + '  diff ' + \
            path.join(directory_name, expected_result_filename) + ' ' + \
            path.join(directory_name, observed_result_filename)

        try:
            observed_result_type = type(observed_result)
            if observed_result_type is OrderedDict:
                self.assertDictEqual(expected_result, observed_result, error_message)
                return
            if observed_result_type is list:
                self.assertListEqual(expected_result, observed_result, error_message)
                return
            self.fail(base_name + ': expected OrderedDict or list, not ' + type(observed_result).__name__)
        except AssertionError:
            self._save_result(expected_result, directory_name, expected_result_filename)
            self._save_result(observed_result, directory_name, observed_result_filename)
            raise

    def assertTextEqual(self, expected_result, observed_result, base_name):

        directory_name = path.join(self._test_dir, 'observed-results')
        expected_result_filename = base_name + '-expected.text'
        observed_result_filename = base_name + '-observed.text'

        error_message = base_name + ': observed value is different than expected:\n' + '  diff ' + \
            path.join(directory_name, expected_result_filename) + ' ' + \
            path.join(directory_name, observed_result_filename)

        try:
            self.assertEqual(expected_result, observed_result, error_message)
        except AssertionError:
            self._save_result(expected_result, directory_name, expected_result_filename)
            self._save_result(observed_result, directory_name, observed_result_filename)
            raise

    # endregion

    # region Protected

    _script_root, _script_name = path.split(__file__)
    _script_root = path.normpath(_script_root)
    _script_name = path.splitext(_script_name)[0]

    _output_dir = None
    _test_dir = path.join(_script_root, _script_name)
    _test_app_root = path.join(_script_root, _script_name, 'source')

    @staticmethod
    def _save_result(result, directory_name, filename):

        if not path.isdir(directory_name):
            mkdir(directory_name)

        if type(result) is not string:
            result = json.dumps(result, ensure_ascii=False, indent=2)
            result = string(result)  # because some Python 2.7 json modules sometimes return str

        with io.open(path.join(directory_name, filename), encoding='utf-8', mode='w', newline='') as ostream:
            ostream.write(result)

    @staticmethod
    def _to_ordered_dict(pairs):
        return OrderedDict(pairs)

    # endregion
    pass


if __name__ == '__main__':
    import unittest.main
    unittest.main()
