#!/usr/bin/env python
# coding=utf-8
#
# Copyright © Splunk, Inc. All Rights Reserved.

from __future__ import absolute_import, division, print_function, unicode_literals
from unittest import TestLoader, TextTestRunner, TestSuite
from argparse import ArgumentParser
from glob import glob, iglob
from os import environ, path
from subprocess import Popen, PIPE

import os
import shutil
import sys
import tarfile

from test_slim import download_archive
from test_case_studies import TestUseCases
from slim.utils.internal import string

test_dir = path.dirname(path.join(path.abspath(__file__)))
project_dir = path.dirname(test_dir)
sys.path.insert(0, project_dir)

assert sys.getdefaultencoding() == 'utf-8'

test_case_studies = path.join(test_dir, 'test_case_studies')
addons = path.join(test_case_studies, 'com.splunk.addon')
composite_apps = path.join(test_dir, 'apps')

print('Python version is {0} on {1}'.format(sys.version.replace('\n', ''), sys.platform), file=sys.stderr)
print('Python executable is', sys.executable, file=sys.stderr)
print('Default encoding is UTF-8', file=sys.stderr)
print('SLIM: test partition', file=sys.stderr)


class CommandTestRunner(TextTestRunner):

    def __init__(self, case_study, keep_files=True, *args, **kwargs):
        TextTestRunner.__init__(self, *args, **kwargs)
        self._case_study = case_study
        self._keep_files = keep_files
        self._files_written = None
        self._use_case = None

    def run(self, suite=None):

        self.stream.writeln('')
        self.stream.writeln('======================================================================')
        self.stream.writeln('Case study: ' + self._case_study)

        test_results = []

        for use_case in os.listdir('use_cases'):

            if use_case.startswith('-'):
                continue

            self._use_case = path.join('use_cases', use_case)

            if not path.isdir(self._use_case):
                continue

            if suite is None:
                test_names = TestLoader().getTestCaseNames(TestUseCases)
                suite = TestSuite()
                for name in test_names:
                    if name in TestUseCases.tests_with_empty_repository_option():
                        suite.addTest(TestUseCases(method_name=name, repository='repository_none'))
                    suite.addTest(TestUseCases(method_name=name, repository='repository'))

            try:
                self.set_up()
                self.stream.writeln('----------------------------------------------------------------------')
                self.stream.writeln(use_case)
                test_result = TextTestRunner.run(self, suite)
                test_results.append(test_result)
            finally:
                self.tear_down()

        return test_results

    def set_up(self):

        self._files_written = []

    def tear_down(self):

        if self._keep_files or self._files_written is None:
            return

        for fn in self._files_written:
            try:
                if path.isdir(fn):
                    shutil.rmtree(fn)
                else:
                    os.remove(fn)
            except EnvironmentError as error:
                raise IOError(error.errno, 'Could not remove {0}: {1}'.format(path.realpath(fn), string(error)), fn)

        self._files_written = []

    def _makeResult(self):
        value = TextTestRunner._makeResult(self)
        value.use_case = self._use_case
        return value


class PartitionTestRunner(TextTestRunner):

    def __init__(self, case_study, keep_files=True, *args, **kwargs):
        TextTestRunner.__init__(self, *args, **kwargs)
        self._case_study = case_study
        self._keep_files = keep_files
        self._removal_list = None

    def run(self, test_suite):

        self.stream.writeln('')
        self.stream.writeln('======================================================================')
        self.stream.writeln('Case study: ' + self._case_study)
        self.stream.writeln('----------------------------------------------------------------------')

        cwd = os.getcwd()

        try:
            os.chdir(self._case_study)

            try:
                self.set_up()
                return TextTestRunner.run(self, test_suite)
            finally:
                self.tear_down()

        finally:
            os.chdir(cwd)

    def set_up(self):

        # Download the case study package before we try to find it below
        download_archive(path.join(project_dir, 'test'))

        filenames = glob('source.tar.gz')
        pattern = None

        if len(filenames) == 0:
            pattern = 'com.splunk.addon-' + self._case_study + '-*.tar.gz'
            filenames = glob(pattern)

        if len(filenames) == 0:
            self.stream.writeln('Expected to find a single source archive matching source.tar.gz or ' + pattern)
            sys.exit(1)

        source_directory = 'com.splunk.addon-' + self._case_study
        source_package = filenames[0]

        if path.isdir(source_directory):
            shutil.rmtree(source_directory)

        if path.isdir('repository'):
            shutil.rmtree('repository')

        with tarfile.open(source_package) as source_package:
            source_package.extractall()

        if not path.isdir(source_directory):
            self.stream.writeln('Failed to extract ' + source_directory + ' from ' + source_package.name)
            self.stream.writeln(' Could this be the result of an archive folder name error?')
            sys.exit(1)

        self._removal_list = [source_directory, 'repository']
        return

    def tear_down(self):

        if self._keep_files or self._removal_list is None:
            return

        for fn in self._removal_list:
            try:
                if path.isdir(fn):
                    shutil.rmtree(fn)
                elif path.isfile(fn):
                    os.remove(fn)
            except EnvironmentError as error:
                raise IOError(error.errno, 'Could not remove {0}: {1}'.format(path.realpath(fn), string(error)), fn)

        self._removal_list = None


splunk_home = string(environ['SPLUNK_HOME'])
splunk_command = path.join(splunk_home, 'bin', 'splunk')
splunk_etc_dir = string(environ['SPLUNK_ETC']) if 'SPLUNK_ETC' in environ else path.join(splunk_home, 'etc')

package_command = (splunk_command, 'cmd', 'python', path.join(project_dir, 'package.py'))


def package(source, repository_path):
    command = package_command + ('--debug', '--output-dir', repository_path, '--repository', repository_path, source)
    process = Popen(command, stdout=PIPE, stderr=PIPE)
    output, diagnostics = process.communicate()
    if process.returncode != 0:
        print(diagnostics)
        print(output)
        sys.exit(process.returncode)
    return


parser = ArgumentParser()

parser.add_argument('--suite', choices=frozenset(('all', 'package', 'partition', 'update-installation')), default='all')
parser.add_argument('--keep', action='store_true', default=False)
args = parser.parse_args(sys.argv[1:])

excluded_addons = ['Splunk_TA_aws', 'splunk_app_aws', 'splunk_app_for_nix', 'splunk_app_okta', 'splunk_app_windows_infrastructure']
current_directory = os.getcwd()
status = 0

try:

    if args.suite in ('all', 'package'):
        # TODO Make this test work
        os.chdir(addons)

        for addon in sorted(os.listdir(addons)):
            test_suite = TestLoader().discover(test_case_studies, pattern='test_package.py')
            if not path.isdir(addon):
                continue
            if addon in excluded_addons:
                continue
            test_runner = PartitionTestRunner(addon, keep_files=args.keep)
            result = test_runner.run(test_suite)
            if result.errors or result.failures:
                status = 1

    if args.suite in ('all', 'partition'):
        os.chdir(addons)

        for addon in sorted(os.listdir(addons)):
            test_suite = TestLoader().discover(test_case_studies, pattern='test_partition.py')
            if not path.isdir(addon):
                continue
            if addon in excluded_addons:
                continue
            test_runner = PartitionTestRunner(addon, keep_files=args.keep)
            result = test_runner.run(test_suite)
            if result.errors or result.failures:
                status = 1

    if args.suite in ('all', 'update-installation'):
        repository_path = os.path.join(composite_apps, 'repository')
        os.chdir(composite_apps)

        if path.isdir(repository_path):
            shutil.rmtree(repository_path)

        if path.isdir('source'):
            shutil.rmtree('source')

        # Download the source package
        download_archive(path.join(project_dir, 'test'))

        with tarfile.open('source.tar.gz') as source:
            source.extractall()

        os.chdir('source')

        for app_root in sorted(os.listdir('.'), reverse=True):
            if not path.isdir(app_root):
                continue
            app_root = path.join(app_root, app_root.split('-', 1)[1])
            manifest = path.join(app_root, 'app.manifest')
            for versioned_manifest in iglob(path.join(app_root, 'app-[0-9].[0-9].[0-9].manifest')):
                shutil.copy(versioned_manifest, manifest)
                package(app_root, repository_path)

        os.chdir('..')

        test_runner = CommandTestRunner(composite_apps, keep_files=args.keep)
        results = test_runner.run()

        if sum(len(result.errors) + len(result.failures) for result in results) > 0:
            status = 1

    if not args.keep:
        if status == 0 and path.isfile('source.tar.gz'):
            shutil.rmtree('repository_none')
            shutil.rmtree('repository')
            shutil.rmtree('source')
            os.remove('source.tar.gz')

    os.chdir(composite_apps)

finally:
    os.chdir(current_directory)

sys.exit(status)
