#!/usr/bin/env python
# coding=utf-8
#
# Copyright © 2015 Splunk, Inc. All Rights Reserved.

from __future__ import absolute_import, division, print_function, unicode_literals

from collections import OrderedDict
from glob import glob
from os import path
import os

from shutil import rmtree
from subprocess import PIPE, Popen
from sys import stderr, stdout

import tarfile

from slim.utils import SlimLogger
from slim.utils.internal import hash_object, string
from slim.app import AppConfiguration

from test_slim import TestCommand


class TestPackage(TestCommand):

    def test_load_configuration(self):

        expected_configuration_json = path.join('expected-results', self._app_under_test + '.configuration.json')
        observed_configuration_json = path.join(self._output_dir, path.basename(expected_configuration_json))

        SlimLogger.reset_counts()
        app_configuration = AppConfiguration.load(self._app_root)
        self.assertEqual(SlimLogger.error_count(), 0)
        app_configuration.save(observed_configuration_json, indent=True)

        observed_result = self._load_observed_result(observed_configuration_json, directory=None)
        expected_result = self._load_expected_result(expected_configuration_json, directory=None)

        self.assertJsonEqual(expected_result, observed_result, path.basename(observed_configuration_json))
        self._removal_list.append(self._output_dir)
        return

    def test_package(self):

        output_dir = self._output_dir

        command = TestPackage._package_command + ('@package.args', '--output-dir', output_dir, self._app_root)
        process = Popen(command, stdout=PIPE, stderr=PIPE, env=os.environ)
        observed_result, observed_diagnostics = process.communicate()
        observed_result = string(observed_result)
        observed_diagnostics = string(observed_diagnostics)
        expected_source_package = path.join(self._app_under_test + '.tar.gz')
        observed_source_package = path.abspath(path.join(output_dir, path.basename(expected_source_package)))
        expected_diagnostics = (
            os.linesep + 'package: [NOTE] Source package exported to "' + observed_source_package + '"' + os.linesep)
        self.assertOutputEndsWith(observed_diagnostics, expected_diagnostics, command, stderr)
        self.assertOutputEndsWith(observed_result, None, command, stdout)
        self.assertTarfileEqual(expected_source_package, observed_source_package, 'Source package')

        self._removal_list.extend((self._case_study, output_dir))
        return

    @classmethod
    def setUpClass(cls):
        super(TestPackage, cls).setUpClass()
        current_directory = os.getcwd()
        cls._case_study = path.basename(current_directory)

    def setUp(self):
        super(TestPackage, self).setUp()
        output_dir = self._test_name
        if path.isdir(output_dir):
            rmtree(output_dir)
        os.mkdir(output_dir)
        self._output_dir = output_dir

    def assertTarfileEqual(self, expected, observed, object_type):

        digests = OrderedDict()

        for source_package in expected, observed:
            digests[source_package] = digest = []
            source = source_package + '.temp'
            if path.isdir(source):
                rmtree(source)
            with tarfile.open(source_package) as tar:
                tar.extractall(source)
            for root, directories, files in os.walk(source):
                root = root[len(source) + 1:]
                for file in sorted(directories):
                    if len(root) > 0:
                        file = path.join(root, file)
                    size = path.getsize(file)
                    digest.append(OrderedDict((
                        ('file', file),
                        ('hash', None),
                        ('size', size),
                        ('type', 'd')
                    )))
                for file in sorted(files):
                    if len(root) > 0:
                        file = path.join(root, file)
                    size = path.getsize(file)
                    digest.append(OrderedDict((
                        ('file', file),
                        ('hash', hash_object(file, size)),
                        ('size', size),
                        ('type', 'f')
                    )))
            rmtree(source)

        try:
            self.assertListEqual(digests[expected], digests[observed])
        except AssertionError:
            failures = self._get_observed_results_directory()
            expected_digest_filename = path.join(failures, path.basename(expected) + '.digest.json')
            self._save_observed_result(digests[expected], expected_digest_filename)
            observed_digest_filename = path.join(failures, path.basename(observed) + '.digest.json')
            self._save_observed_result(digests[observed], observed_digest_filename)
            self.fail(
                object_type + ': observed tarfile is different than expected:\n' + '  diff ' +
                expected_digest_filename + ' ' + observed_digest_filename)

        return

    _package_command = (
        TestCommand._splunk_command, 'cmd', 'python', path.join(TestCommand._project_root, 'package.py'))
