#!/usr/bin/env python
# coding=utf-8
#
# Copyright © 2015 Splunk, Inc. All Rights Reserved.

from __future__ import absolute_import, division, print_function, unicode_literals

import fnmatch

from glob import iglob

from os import path
import os

from shutil import rmtree
from subprocess import PIPE, Popen
from sys import stderr, stdout

from slim.utils.internal import string
from test_slim import TestCommand


class TestPartition(TestCommand):

    def test_partition(self):

        output_dir = 'partition.0.output'

        if path.isdir(output_dir):
            rmtree(output_dir)

        command = TestPartition._partition + ('@partition.0.args', '--output-dir', output_dir, self._source_package)
        process = Popen(command, stdout=PIPE, stderr=PIPE, env=os.environ)

        observed_result, observed_diagnostics = process.communicate()
        observed_result = string(observed_result)
        observed_diagnostics = string(observed_diagnostics)

        pattern = path.join('expected-results', self._app_under_test + '-*.file-digest.json')

        expected_deployment_packages = [
            path.join(output_dir, path.basename(name[:-len('file-digest.json')]) + 'tar.gz') for name in iglob(pattern)
        ]

        group, name, version = self._app_under_test.split('-')
        app_id = group + '-' + name + ':' + version

        expected_diagnostics = (
            '{nl}partition: [NOTE] Generated deployment packages for {app_id}:{nl}  {deployment_packages}{nl}'.format(
                nl=os.linesep,
                app_id=app_id,
                deployment_packages=(os.linesep + '  ').join(
                    sorted((path.abspath(package) for package in expected_deployment_packages), reverse=True)
                )
            )
        )

        self.assertOutputEndsWith(observed_diagnostics, expected_diagnostics, command, stderr)
        self.assertOutputEndsWith(observed_result, None, command, stdout)

        filename_prefixes = [name[:-len('tar.gz')] for name in expected_deployment_packages]

        for suffix in 'configuration.json', 'file-digest.json':
            count = None
            for count, name in enumerate((name + suffix for name in filename_prefixes), 1):
                expected_configuration = self._load_expected_result(name, directory='expected-results')
                observed_configuration = self._load_observed_result(name, directory=None)
                self.assertJsonEqual(expected_configuration, observed_configuration, path.basename(name))
            self.assertEqual(count, len(filename_prefixes))

        self._removal_list.append(output_dir)

    def test_partition_combined_search_head_indexer_workloads(self):

        output_dir = 'partition.1.output'

        if path.isdir(output_dir):
            rmtree(output_dir)

        command = TestPartition._partition + ('@partition.1.args', '--output-dir', output_dir, self._source_package)
        process = Popen(command, stdout=PIPE, stderr=PIPE)
        observed_result, observed_diagnostics = process.communicate()

        self.assertProcessReturnCodeEqual(0, process, command, stderr, observed_diagnostics)

        expected = ['*' + sc + '.tar.gz' for sc in ('_forwarders', '_search_head_indexers')]
        expected.append('installation-update.json')
        expected.append('installation-actions.json')

        actual = os.listdir(output_dir)

        for e in expected:
            filtered = len(fnmatch.filter(actual, e))
            self.assertNotEqual(filtered, 0)

        self._removal_list.append(output_dir)

    def test_partition_only(self):

        command = TestPartition._partition + ('--repository', 'repository') + ('@partition.3.args',)
        output_dir = 'partition.3.output'

        if path.isdir(output_dir):
            rmtree(output_dir)

        process = Popen(command, stdout=PIPE, stderr=PIPE)
        observed_result, observed_diagnostics = process.communicate()

        try:
            self.assertEqual(process.returncode, 0)
        except AssertionError as error:
            self.fail(string(error) + '\nCommand: ' + ' '.join(command) + '\nDiagnostics:\n' + observed_diagnostics)

        expected = ['*' + sc + '.tar.gz' for sc in ('forwarder', 'indexer', 'search-head')]
        expected.append('installation-actions.json')

        not_expected = ['installation-update.json']

        actual = os.listdir(output_dir)

        for e in expected:
            filtered = len(fnmatch.filter(actual, e))
            self.assertNotEqual(filtered, 0)

        for n in not_expected:
            filtered = len(fnmatch.filter(actual, n))
            self.assertEqual(filtered, 0)

        # self.assertListEqual(sorted(expected), sorted(os.listdir(output_dir)))  # because sort order varies by platform

        self._removal_list.append(output_dir)

    def test_partition_separate_search_head_indexer_workloads(self):

        output_dir = 'partition.2.output'

        if path.isdir(output_dir):
            rmtree(output_dir)

        command = TestPartition._partition + ('@partition.2.args', '--output-dir', output_dir, self._source_package)
        process = Popen(command, stdout=PIPE, stderr=PIPE)
        observed_result, observed_diagnostics = process.communicate()

        self.assertProcessReturnCodeEqual(0, process, command, stderr, observed_diagnostics)

        expected = ['*' + sc + '.tar.gz' for sc in ('_forwarders', '_indexers', '_search_heads')]
        expected.append('installation-update.json')
        expected.append('installation-actions.json')

        actual = os.listdir(output_dir)

        for e in expected:
            filtered = len(fnmatch.filter(actual, e))
            self.assertNotEqual(filtered, 0)

        self._removal_list.append(output_dir)

    _partition = (
        TestCommand._splunk_command, 'cmd', 'python', path.join(TestCommand._project_root, 'partition.py'))
