#!/usr/bin/env python
# coding=utf-8
#
# Copyright © Splunk, Inc. All Rights Reserved.

from __future__ import absolute_import, division, print_function, unicode_literals

from glob import glob
from os import getcwd, path
from subprocess import PIPE, Popen
from tempfile import mkdtemp

from slim.utils import encode_filename, encode_string
from slim.utils.internal import string

from test_slim import TestCommand

import io


class TestUseCases(TestCommand):

    def __init__(self, method_name, repository):
        super(TestUseCases, self).__init__(method_name)
        self._command_args = None
        self._repository = repository

    def run(self, result=None):
        self.assertTrue(hasattr(result, 'use_case'))
        self._use_case = result.use_case
        self._use_case_number = path.basename(self._use_case).split('.')[0]
        TestCommand.run(self, result)

    def setUp(self):
        command_file = path.join(self._use_case, self._testMethodName[len('test_'):] + '.args')
        if not path.isfile(command_file):
            self.skipTest('Undefined for this use case: ' + self._use_case)
        self._command_args = ('--repository', path.join(getcwd(), self._repository), '@' + command_file)
        self._output_dir = mkdtemp(prefix='slim-output_dir-', dir=path.abspath(path.join('..', '..', '..', '..')))
        self._removal_list = []
        return

    def test_describe(self):

        filename_expected = path.join(self._use_case, 'describe.app.text')
        filename_observed = path.join(self._output_dir, 'describe.app.text')
        command = TestUseCases._describe + self._command_args + ('--output', filename_observed)

        process, observed_result, observed_diagnostics = self._execute(command)

        self.assertEqual(process.returncode, 0)
        self.assertEqual(len(observed_result), 0, 'expected zero-length result, not:\n' + observed_result)

        expected_diagnostics = 'describe: Describing "repository'

        self.assertTrue(
            observed_diagnostics.startswith(expected_diagnostics), 'expected a diagnostic message starting with \'' +
            expected_diagnostics + '\', not:\n' + observed_diagnostics)

        filename = filename_expected

        try:
            expected = TestUseCases._load_result(filename)
            filename = filename_observed
            observed = TestUseCases._load_result(filename)
        except IOError as error:
            self.fail(
                'Could not load ' + encode_filename(filename) + ': ' + string(error) + '\nDiagnostics: ' +
                observed_diagnostics)

        self.assertEqual(expected, observed)
        self._removal_list.append(self._output_dir)

    def test_install(self):

        output_dir = self._output_dir
        command = TestUseCases._update_installation + self._command_args + ('--output-dir', output_dir)
        process, observed_result, observed_diagnostics = self._execute(command)

        # TODO: Dnoble: incorporate pass/fail logic (e.g., by introducing install.{passed,failed} files used in a
        # parallel test (e.g., test_failed_install)(?))
        # self.assertProcessReturnCodeEqual(0, process, command, stderr, observed_diagnostics)

        check_file_list = [
            (
                path.join(self._use_case, 'install.installation-update.expected.json'),
                path.join(output_dir, 'installation-update.json')
            ),
            (
                path.join(self._use_case, 'install.graph-updates.expected.json'),
                path.join(output_dir, 'graph-updates.json')
            )
        ]

        for (filename_expected, filename_observed) in check_file_list:
            expected_installation_update = path.exists(filename_expected)
            found_installation_update = path.exists(filename_observed)

            if expected_installation_update and not found_installation_update:
                self.fail(
                    'Expected to find ' + encode_string(filename_observed) + '\n  command: ' + ' '.join(command) +
                    '\n  diagnostics: ' + observed_diagnostics
                )

            if not expected_installation_update and found_installation_update:
                self.fail(
                    'Did not expect to find ' + encode_string(filename_observed) + '\n  command: ' + ' '.join(command) +
                    '\n  diagnostics: ' + observed_diagnostics
                )

            if expected_installation_update and found_installation_update:
                filename = filename_expected
                try:
                    expected = self._load_result(filename) % {'output_dir': output_dir.replace('\\', '\\\\')}
                    filename = filename_observed
                    observed = TestUseCases._load_result(filename)
                except IOError as error:
                    self.fail(
                        'Could not load ' + encode_string(filename) + ': ' + string(error) + '\n  diagnostics: ' +
                        observed_diagnostics)
                self.assertJsonEqual(observed, expected, path.basename(self._use_case) + '-' + path.basename(filename_observed))

        self._removal_list.append(output_dir)
        return

    def test_package(self):
        pass  # TODO: implement TestUseCases.test_package and then cover them in test/apps/use_cases

    def test_partition(self):

        output_dir = self._output_dir
        command = TestUseCases._partition + self._command_args + ('--output-dir', output_dir)
        process, observed_result, observed_diagnostics = self._execute(command)

        if len(observed_result) > 0:
            self.fail('No result expected from command: ' + ' '.join(command) + '\n' + observed_result)

        # On partition failure we will not have a new installation graph - no compare needed in this case

        output_dir_replacement = path.join(output_dir, '').replace('\\', '\\\\')

        for name in 'update', 'actions':
            filename_expected = path.join(self._use_case, 'partition.installation-' + name + '.expected.json')
            expected_installation_update = path.exists(filename_expected)

            filename_observed = path.join(output_dir, 'installation-' + name + '.json')
            found_installation_update = path.exists(filename_observed)

            if expected_installation_update and not found_installation_update:
                self.fail(
                    'Expected to find ' + encode_string(filename_observed) + '\n  command: ' + ' '.join(command) +
                    '\n  diagnostics: ' + observed_diagnostics
                )

            if not expected_installation_update and found_installation_update:
                self.fail(
                    'Did not expect to find ' + encode_string(filename_observed) + '\n  command: ' + ' '.join(command) +
                    '\n  diagnostics: ' + observed_diagnostics
                )

            if expected_installation_update and found_installation_update:
                filename = filename_expected
                try:
                    expected = TestUseCases._load_result(filename) % {'output_dir': output_dir_replacement}
                    filename = filename_observed
                    observed = TestUseCases._load_result(filename)
                except IOError as error:
                    self.fail(
                        'Could not load ' + encode_string(filename) + ': ' + string(error) + '\n  diagnostics: ' +
                        observed_diagnostics)
                self.assertJsonEqual(expected, observed, path.basename(self._use_case) + '-installation_' + name + '.json')

        partition_expected = path.join(self._use_case, 'partition.expected')

        if path.isdir(partition_expected):
            path_sep = '\\\\' if path.sep == '\\' else path.sep
            for pattern in '*.configuration.json', '*.file-digest.json':
                for filename_expected in sorted(glob(path.join(partition_expected, pattern))):
                    filename_observed = path.join(output_dir, path.basename(filename_expected))
                    filename = filename_expected
                    try:
                        expected = TestUseCases._load_result(filename) % {'path.sep': path_sep}
                        filename = filename_observed
                        observed = TestUseCases._load_result(filename)
                    except IOError as error:
                        self.fail(
                            'Could not load ' + encode_string(filename) + ': ' + string(error) +
                            '\n  command: ' + ' '.join(command) + '\n  diagnostics: ' +
                            observed_diagnostics
                        )
                    self.assertJsonEqual(expected, observed, path.basename(filename))

        self._removal_list.append(output_dir)

    def test_uninstall(self):

        output_dir = self._output_dir
        command = TestUseCases._update_installation + self._command_args + ('--output-dir', output_dir)
        process, observed_result, observed_diagnostics = self._execute(command)

        error_message = 'No diagnostics from command: ' + ' '.join(command) + '\n' + observed_diagnostics
        self.assertGreater(len(observed_diagnostics), 0, error_message)

        error_message = 'No result expected from command: ' + ' '.join(command) + '\n' + observed_result
        self.assertEqual(len(observed_result), 0, error_message)

        # On uninstall failure we will not have a new installation graph - no compare needed in this case

        check_file_list = [
            (path.join(self._use_case, 'uninstall.installation-update.expected.json'), path.join(output_dir, 'installation-update.json')),
            (path.join(self._use_case, 'uninstall.graph-updates.expected.json'), path.join(output_dir, 'graph-updates.json'))
        ]

        for (filename_expected, filename_observed) in check_file_list:
            expected_installation_update = path.exists(filename_expected)
            found_installation_update = path.exists(filename_observed)

            if expected_installation_update and not found_installation_update:
                self.fail('Expected to find ' + encode_string(filename_observed) + '\nDiagnostics: ' + observed_diagnostics)

            elif not expected_installation_update and found_installation_update:
                self.fail('Did not expected to find ' + encode_string(filename_observed) + '\nDiagnostics: ' + observed_diagnostics)

            elif expected_installation_update and found_installation_update:
                filename = filename_expected
                try:
                    expected = TestUseCases._load_result(filename)
                    filename = filename_observed
                    observed = TestUseCases._load_result(filename)
                except IOError as error:
                    self.fail('Could not load ' + encode_string(filename) + ': ' + string(error) + '\nDiagnostics: ' + observed_diagnostics)
                self.assertJsonEqual(expected, observed, path.basename(self._use_case) + '-' + path.basename(filename_observed))

        self._removal_list.append(output_dir)

    def test_update(self):

        output_dir = self._output_dir
        command = TestUseCases._update_installation + self._command_args + ('--output-dir', output_dir)
        process, observed_result, observed_diagnostics = self._execute(command)

        error_message = 'No diagnostics from command: ' + ' '.join(command) + '\n' + observed_diagnostics
        self.assertGreater(len(observed_diagnostics), 0, error_message)

        error_message = 'No result expected from command: ' + ' '.join(command) + '\n' + observed_result
        self.assertEqual(len(observed_result), 0, error_message)

        check_file_list = [
            (path.join(self._use_case, 'update.installation-update.expected.json'), path.join(output_dir, 'installation-update.json')),
            (path.join(self._use_case, 'update.graph-updates.expected.json'), path.join(output_dir, 'graph-updates.json'))
        ]

        for (filename_expected, filename_observed) in check_file_list:

            if not path.exists(filename_observed):
                self.fail('Expected to find ' + encode_string(filename_observed) + '\nDiagnostics: ' + observed_diagnostics)

            filename = filename_expected
            try:
                expected = TestUseCases._load_result(filename)
                filename = filename_observed
                observed = TestUseCases._load_result(filename)
            except IOError as error:
                self.fail('Could not load ' + encode_string(filename) + ': ' + string(error) + '\nDiagnostics: ' + observed_diagnostics)
            self.assertJsonEqual(expected, observed, path.basename(self._use_case) + '-' + path.basename(filename_observed))

        self._removal_list.append(output_dir)

    @staticmethod
    def _load_result(filename):
        with io.open(path.join(filename), encoding='utf-8') as f:
            return f.read()

    @staticmethod
    def _execute(command):
        process = Popen(command, stdout=PIPE, stderr=PIPE)
        result, diagnostics = process.communicate()
        return process, string(result), string(diagnostics)

    @staticmethod
    def tests_with_empty_repository_option():
        return ["test_describe"]

    _describe = (
        TestCommand._splunk_command, 'cmd', 'python', path.join(TestCommand._project_root, 'describe.py'), '--debug')

    _package = (
        TestCommand._splunk_command, 'cmd', 'python', path.join(TestCommand._project_root, 'package.py'), '--debug')

    _partition = (
        TestCommand._splunk_command, 'cmd', 'python', path.join(TestCommand._project_root, 'partition.py'), '--debug')

    _update_installation = (
        TestCommand._splunk_command, 'cmd', 'python', path.join(TestCommand._project_root, 'update_installation.py'),
        '--debug')
