#!/usr/bin/env python
# coding=utf-8
#
# Copyright © Splunk, Inc. All Rights Reserved.

from __future__ import absolute_import, division, print_function, unicode_literals
from unittest import TestLoader, TextTestRunner
from argparse import ArgumentParser
from os import path
from shutil import rmtree
from subprocess import Popen, PIPE

import json
import os
import shutil
import sys
import tarfile

from slim.app import AppConfiguration
from slim.utils.internal import string
from test_slim import TestCommand

test_dir = path.dirname(path.join(path.abspath(__file__)))
project_dir = path.dirname(test_dir)
sys.path.insert(0, project_dir)

assert sys.getdefaultencoding() == 'utf-8'

addons = path.join(test_dir, 'test_generate_manifest')

print('Python version is {0} on {1}'.format(sys.version.replace('\n', ''), sys.platform), file=sys.stderr)
print('Python executable is', sys.executable, file=sys.stderr)
print('Default encoding is UTF-8', file=sys.stderr)
print('SLIM: test generate-manifest', file=sys.stderr)


class GenerateManifestTestRunner(TextTestRunner):

    def __init__(self, test_generate_manifest, keep_files=True, *args, **kwargs):
        TextTestRunner.__init__(self, *args, **kwargs)
        self._test_app = test_generate_manifest
        self._keep_files = keep_files
        self._files_written = None

    def run(self, test):

        self.stream.writeln('')
        self.stream.writeln('----------------------------------------------------------------------')
        self.stream.writeln('Test slim generate-manifest: ' + self._test_app)
        self.stream.writeln('----------------------------------------------------------------------')

        cwd = os.getcwd()

        try:
            os.chdir(self._test_app)

            try:
                self.set_up()
                return TextTestRunner.run(self, test)
            finally:
                self.tear_down()

        finally:
            os.chdir(cwd)

    def set_up(self):
        self._files_written = []
        return

    def tear_down(self):

        if self._keep_files or self._files_written is None:
            return

        for fn in self._files_written:
            try:
                if path.isdir(fn):
                    shutil.rmtree(fn)
                else:
                    os.remove(fn)
            except EnvironmentError as error:
                raise IOError(error.errno, 'Could not remove {0}: {1}'.format(path.realpath(fn), string(error)), fn)

        self._files_written = []


class TestGenerateManifest(TestCommand):

    def test_load_configuration(self):

        expected_configuration_json = path.join('expected-results', self._app_under_test + '.configuration.json')
        expected_result = self._load_expected_result(expected_configuration_json, directory=None)

        observed_configuration_json = path.join(self._test_name, self._app_under_test + '.configuration.json')
        app_configuration = AppConfiguration.load(self._app_root)
        app_configuration.save(observed_configuration_json, indent=True)
        observed_result = self._load_observed_result(expected_configuration_json, directory=self._test_name)

        self.assertJsonEqual(expected_result, observed_result, path.basename(observed_configuration_json))
        self._removal_list.extend([self._app_root, self._test_name])
        return

    def test_generate_manifest(self):

        command = TestGenerateManifest._generate_manifest + (self._app_root,)
        process = Popen(command, stdout=PIPE, stderr=PIPE)
        expected_result = self._load_expected_result('app.manifest.json', directory='expected-results')

        self.assertProcessResultEqual(expected_result, process, 'app.manifest')
        self._removal_list.extend([self._app_root, self._test_name])
        return

    def assertProcessResultEqual(self, expected, process, basename):

        observed = self._get_result_from_process(process)
        expected = json.loads(expected, object_pairs_hook=TestCommand._to_ordered_dict)

        self.assertTrue('diagnostics' in expected, 'Missing diagnostics field in expected results for ' + basename)
        self.assertTrue('result' in expected, 'Missing result field in expected results for ' + basename)
        self.assertEqual(len(expected), 2)
        self.assertJsonEqual(expected, observed, basename)

        return

    def setUp(self):
        super(TestGenerateManifest, self).setUp()
        if path.isdir(self._app_root):
            rmtree(self._app_root)
        source_package = path.join(self._app_under_test + '.tar.gz')
        with tarfile.open(source_package) as tar:
            tar.extractall()
        if path.isdir(self._test_name):
            rmtree(self._test_name)
        os.mkdir(self._test_name)
        return

    _generate_manifest = (
        TestCommand._splunk_command, 'cmd', 'python', path.join(TestCommand._project_root, 'generate_manifest.py'),
        '--debug')

parser = ArgumentParser()

parser.add_argument('--keep', action='store_true', default=False)
args = parser.parse_args(sys.argv[1:])

current_directory = os.getcwd()
status = 0

try:
    os.chdir(addons)
    for test_app in sorted(os.listdir(addons)):
        test_suite = TestLoader().loadTestsFromModule(sys.modules[__name__])
        if not path.isdir(test_app):
            continue
        test_runner = GenerateManifestTestRunner(test_app, keep_files=args.keep)
        result = test_runner.run(test_suite)
        if result.errors or result.failures:
            status = 1
finally:
    os.chdir(current_directory)

sys.exit(status)
