#!/usr/bin/env python
# coding=utf-8
#
# Copyright © 2015 Splunk, Inc. All Rights Reserved.

from __future__ import absolute_import, division, print_function, unicode_literals

from unittest import TestCase
from collections import OrderedDict
from glob import glob
from subprocess import Popen, PIPE

import io
import json

from os import environ, getcwd, mkdir, path, remove
from shutil import rmtree
from tempfile import mkdtemp

from slim.utils import slim_configuration
from slim.utils.internal import string


__all__ = ['TestCommand', 'download_archive']


class TestCommand(TestCase):

    def __init__(self, method_name):

        TestCase.__init__(self, method_name)

        self._test_name = type(self).__name__ + '.' + self._testMethodName
        self._app_root = None
        self._app_under_test = None
        self._removal_list = None
        self._repository_path = None
        self._source_package = None
        self._temp_directory_path = None

        self.maxDiff = None

    def setUp(self):

        current_directory = getcwd()
        name = path.basename(current_directory)
        group = path.basename(path.dirname(current_directory))
        patterns = [app_id + '-[0-9].[0-9].[0-9].tar.gz' for app_id in (name, group + '-' + name)]

        download_archive(path.join(self._project_root, 'test'))
        source_packages = []

        for pattern in patterns:
            source_packages = glob(pattern)
            if len(source_packages) > 0:
                break

        self.assertNotEqual(len(source_packages), 0, 'Matching source package not found: ' + string(patterns))

        self.assertEqual(
            len(source_packages), 1, 'Expected 1 source package matching ' + string(patterns) + ', not ' +
            string(len(source_packages)) + ': ' + string(source_packages))

        self._app_root = group + '-' + name
        self._source_package = source_packages[0]
        self._app_under_test = path.basename(self._source_package[:-len('.tar.gz')])

        repository_path = mkdtemp(
            prefix='-'.join(('slim', 'repository', '')),
            dir=TestCommand._splunk_source)

        temp_directory_path = mkdtemp(
            prefix='-'.join(('slim', 'temp', '')),
            dir=TestCommand._splunk_source)

        self._repository_path = repository_path
        self._temp_directory_path = temp_directory_path
        self._removal_list = [repository_path, temp_directory_path, self._source_package]

    def tearDown(self):

        for filename in self._removal_list:
            if path.isdir(filename):
                rmtree(filename)
            elif path.isfile(filename) or path.islink(filename):
                remove(filename)

    def assertJsonEqual(self, expected_result, observed_result, base_name):

        filename = self._test_name if base_name is None else self._test_name + '-' + base_name
        directory = self._get_observed_results_directory()

        expected_result_filename = path.join(directory, filename + '-expected.json')
        observed_result_filename = path.join(directory, filename + '-observed.json')

        try:
            if type(observed_result) is string:
                try:
                    observed_result = json.loads(observed_result, object_pairs_hook=TestCommand._to_ordered_dict)
                except ValueError as error:
                    raise ValueError(base_name + ': Could not load observed_result from string: ' + string(error))

            if type(expected_result) is string:
                try:
                    expected_result = json.loads(expected_result, object_pairs_hook=TestCommand._to_ordered_dict)
                except ValueError as error:
                    raise ValueError(base_name + ': Could not load expected_result from string: ' + string(error))

        except ValueError as error:
            self._save_observed_result(expected_result, expected_result_filename)
            self._save_observed_result(observed_result, observed_result_filename)
            self.fail(
                string(error) +
                '\n  diff ' + expected_result_filename + ' ' + observed_result_filename +
                '\n  diff ' + path.abspath(path.join('expected-results', base_name)) + ' ' + observed_result_filename
            )

        error_message = (
            'JSON document is different than expected:' +
            '\n  diff ' + expected_result_filename + ' ' + observed_result_filename +
            '\n  diff ' + path.abspath(path.join('expected-results', base_name)) + ' ' + observed_result_filename
        )

        try:
            observed_result_type = type(observed_result)
            if observed_result_type is OrderedDict:
                self.assertDictEqual(expected_result, observed_result, error_message)
                return
            if observed_result_type is list:
                self.assertListEqual(expected_result, observed_result, error_message)
                return
            self.fail(base_name + ': expected OrderedDict or list, not ' + type(observed_result).__name__)
        except AssertionError:
            self._save_observed_result(expected_result, expected_result_filename)
            self._save_observed_result(observed_result, observed_result_filename)
            raise

    def assertOutputEndsWith(self, observed, expected, command, device):
        condition = len(observed) == 0 if expected is None else observed.endswith(expected)
        if condition is False:
            self.fail('Unexpected command output:\nCommand:\n  {command}\nOutput on {device}:\n  {output}'.format(
                command=' '.join(command), device=device.name, output=observed.replace('\n', '\n  ')
            ))

    def assertProcessReturnCodeEqual(self, expected, process, command, device, diagnostics):
        if process.returncode != expected:
            self.fail(
                'Expected process.returncode=0, not process.returncode={returncode}\nCommand: {command}\nOutput on '
                '{device}:\n  {output}'.format(
                    returncode=process.returncode, command=' '.join(command), device=device.name,
                    output=diagnostics.replace('\n', '\n  ')
                )
            )

    @staticmethod
    def _get_result_from_process(process):
        import re
        result, diagnostics = process.communicate()
        text = ''.join(line for line in string(result).splitlines() if re.match(r'\s*#', line) is None)
        # noinspection PyBroadException
        try:
            result_json = json.loads(text, object_pairs_hook=TestCommand._to_ordered_dict)
        except:
            result_json = {}
        return OrderedDict((
            ('diagnostics', string(diagnostics).replace('\r\n', '\n')),
            ('result', result_json)
        ))

    @staticmethod
    def _load_expected_result(filename, directory):
        if directory is not None:
            filename = path.join(directory, path.basename(filename))
            project_cache = directory
        else:
            project_cache = slim_configuration.cache.cache_prefix
        try:
            with io.open(filename, encoding='utf-8', newline=None) as f:
                result = f.read()
        except (IOError, TypeError) as error:
            raise type(error)('Cannot load expected result from ' + filename + ': ' + string(error))
        try:
            result = result % {
                'project_root': getcwd().replace('\\', '\\\\'),
                'path.sep': path.sep.replace('\\', '\\\\'),
                'project_cache': project_cache.replace('\\', '\\\\')
            }
        except KeyError as error:
            raise ValueError(
                'Cannot expand expected result loaded from ' + filename + ': Key ' + string(error) + ' is undefined')

        return result

    @staticmethod
    def _load_observed_result(filename, directory):
        if directory is not None:
            filename = path.join(directory, path.basename(filename))
        try:
            with io.open(filename, encoding='utf-8') as f:
                result = f.read()
        except IOError as error:
            raise type(error)('Cannot load observed result from ' + filename + ': ' + string(error))
        return result

    def _save_observed_result(self, observed_result, filename):

        directory_name = self._get_observed_results_directory()

        if not path.isdir(directory_name):
            mkdir(directory_name)

        if type(observed_result) is not string:
            observed_result = json.dumps(observed_result, ensure_ascii=False, indent=2)
            observed_result = string(observed_result)  # because some Python 2.7 json modules sometimes return str

        with io.open(path.join(directory_name, filename), encoding='utf-8', mode='w', newline='') as f:
            f.write(observed_result)

    @staticmethod
    def _get_observed_results_directory():
        return path.abspath('observed-results')

    @staticmethod
    def _to_ordered_dict(pairs):
        return OrderedDict(pairs)

    _project_root = path.dirname(path.dirname(path.dirname(path.realpath(__file__))))

    _splunk_source = string(environ['SPLUNK_SOURCE'])
    _splunk_home = string(environ['SPLUNK_HOME'])
    _splunk_command = path.join(_splunk_home, 'bin', 'splunk')
    _splunk_etc_dir = string(environ['SPLUNK_ETC']) if 'SPLUNK_ETC' in environ else path.join(_splunk_home, 'etc')


def download_archive(test_dir, download_file='@download.args'):
    command = ('python', path.join(test_dir, 'download_archive.py'), download_file)
    process = Popen(command, stdout=PIPE, stderr=PIPE)
    result, diagnostics = process.communicate()
    if process.returncode != 0:
        print(diagnostics)
        print(result)
