#!/usr/bin/env python
# coding=utf-8
#
# Copyright © Splunk, Inc. All Rights Reserved.

""" The tests in this module make this assumption: The test is

    * Run by the Splunk build system or

    * Started in a dev build environment using this command which ensures that environment variables are set correctly:
      .. code-block::
            splunk cmd python test-ignore.py

The tests in this module depend on these environment variables being set the way `splunk cmd` sets them up:

    * `HOME`
    * `PYTHONPATH`
    * `SPLUNK_HOME`
    * `SPLUNK_SOURCE`

"""

from __future__ import absolute_import, division, print_function, unicode_literals

from slim.utils._configuration import SlimConfigurationManager
from slim.utils.internal import *
from slim.utils import SlimIgnore

from collections import OrderedDict
from os import chdir, environ, getcwd, makedirs, mkdir, path, walk
from shutil import rmtree
from tempfile import mkdtemp
from unittest import TestCase

import io


class TestSlimIgnore(TestCase):

    # region Tests

    def test_anchored_directory(self):
        """ Tests anchored directory patterns

        The anchored directory pattern has two forms:
        .. code-block::
            /b/**
            /c/

        This test runs against the contents of a fictional source directory.

        """
        filenames = OrderedDict([('a', 'd'), ('x/a/x', 'f'), ('x/x/a', 'd')])

        test_variant, working_directory_path = self._prepare_source_directory(filenames)
        errors = []

        try:
            # Variant 1

            errors.extend(
                test_variant(
                    '1.1. Exclude pattern', patterns=['/a/'],
                    expected=['x', 'x/a', 'x/x', 'x/a/x', 'x/x/a']))
            errors.extend(
                test_variant(
                    '1.2. Include pattern', patterns=['!/a/'],
                    expected=['a', 'x', 'x/a', 'x/x', 'x/a/x', 'x/x/a']))
            errors.extend(
                test_variant(
                    '1.3. Exclude, include pattern', patterns=['/a/', '!/a/'],
                    expected=['a', 'x', 'x/a', 'x/x', 'x/a/x', 'x/x/a']))
            errors.extend(
                test_variant(
                    '1.4. Include, exclude pattern', patterns=['!/a/', '/a/'],
                    expected=['a', 'x', 'x/a', 'x/x', 'x/a/x', 'x/x/a']))
            errors.extend(
                test_variant(
                    '1.5. Exclude, include pattern', patterns=['*', '!/a/'], expected=['a']))
            errors.extend(
                test_variant(
                    '1.6. Include, exclude pattern', patterns=['!/a/', '*'], expected=['a']))

            # Variant 2

            errors.extend(
                test_variant(
                    '2.1. Exclude pattern', patterns=['/a/**'],
                    expected=['x', 'x/a', 'x/x', 'x/a/x', 'x/x/a']))
            errors.extend(
                test_variant(
                    '2.2. Include pattern', patterns=['!/a/**'],
                    expected=['a', 'x', 'x/a', 'x/x', 'x/a/x', 'x/x/a']))
            errors.extend(
                test_variant(
                    '2.3. Exclude, include pattern', patterns=['/a/**', '!/a/'],
                    expected=['a', 'x', 'x/a', 'x/x', 'x/a/x', 'x/x/a']))
            errors.extend(
                test_variant(
                    '2.4. Include, exclude pattern', patterns=['!/a/', '/a/**'],
                    expected=['a', 'x', 'x/a', 'x/x', 'x/a/x', 'x/x/a']))
            errors.extend(
                test_variant(
                    '2.5. Exclude, include pattern', patterns=['*', '!/a/**'],
                    expected=['a']))
            errors.extend(
                test_variant(
                    '2.6. Include, exclude pattern', patterns=['!/a/**', '*'],
                    expected=['a']))
        finally:
            chdir(working_directory_path)

        self.assertEqual(len(errors), 0, '\n\n' + '\n\n'.join(errors))

    def test_anchored_file(self):
        """ Tests anchored file patterns

        The anchored file pattern has a single form:
        .. code-block::
            /a

        This test runs against the contents of a fictional source directory.

        """
        filenames = OrderedDict([('a', 'f'), ('x/a', 'f'), ('x/x/a/x', 'f'), ('x/x/x/a', 'd')])
        test_variant, working_directory_path = self._prepare_source_directory(filenames)
        filenames = filenames.keys()
        errors = []

        try:
            errors.extend(
                test_variant(
                    '1. Exclude pattern', patterns=['/a'],
                    expected= ['x', 'x/x', 'x/a', 'x/x/a', 'x/x/x', 'x/x/a/x', 'x/x/x/a']))
            errors.extend(
                test_variant(
                    '2. Include pattern', patterns=['!/a'],
                    expected=['x', 'a', 'x/x', 'x/a', 'x/x/a', 'x/x/x', 'x/x/a/x', 'x/x/x/a']))
            errors.extend(
                test_variant(
                    '3. Exclude, include pattern', patterns=['/a', '!/a'],
                    expected=['x', 'a', 'x/x', 'x/a', 'x/x/a', 'x/x/x', 'x/x/a/x', 'x/x/x/a']))
            errors.extend(
                test_variant(
                    '4. Include, exclude pattern', patterns=['!/a', '/a'],
                    expected=['x', 'a', 'x/x', 'x/a', 'x/x/a', 'x/x/x', 'x/x/a/x', 'x/x/x/a']))
            errors.extend(
                test_variant(
                    '5. Include, exclude pattern', patterns=['!/*', '/a'],
                    expected=['a', 'x', 'x/x', 'x/a', 'x/x/a', 'x/x/x', 'x/x/a/x', 'x/x/x/a']))
            errors.extend(
                test_variant(
                    '6. Include, exclude pattern', patterns=['/a', '!/*'],
                    expected=['a', 'x', 'x/x', 'x/a', 'x/x/a', 'x/x/x', 'x/x/a/x', 'x/x/x/a']))
            errors.extend(
                # Excludes all files and directories at the root, excluding file or directory 'a' at the root
                test_variant(
                    '7. Include, exclude pattern', patterns=['/*', '!/a'],
                    expected=[u'a', u'x/a', u'x/x', u'x/x/a', u'x/x/a/x', u'x/x/x', u'x/x/x/a']))
            errors.extend(
                # Excludes all files and directories at the root, excluding file or directory 'a' at the root
                test_variant(
                    '8. Include, exclude pattern', patterns=['!/a', '/*'],
                    expected=[u'a', u'x/a', u'x/x', u'x/x/a', u'x/x/a/x', u'x/x/x', u'x/x/x/a']))
        finally:
            chdir(working_directory_path)

        self.assertEqual(len(errors), 0, '\n\n' + '\n\n'.join(errors))

    def test_unanchored_directory(self):
        """ Tests three forms of the unanchored directory pattern

        These are the variants tested:
        .. code-block::
            **/a/
            **/a/**
            **/a/b/c

        This test runs against the contents of a fictional source directory.

        """
        filenames = OrderedDict([('a/b/c', 'f'), ('x/a/b', 'f'), ('x/x/a/b', 'f'), ('x/x/x/a/b/c', 'f')])
        test_variant, working_directory_path = self._prepare_source_directory(filenames)
        errors = []

        try:
            errors.extend(
                test_variant(
                    '1.1. Exclude pattern', patterns=['a/'],
                    expected=['x', 'a/b/c', 'x/x', 'x/a/b', 'x/x/x', 'x/x/a/b', 'x/x/x/a/b/c']))
            errors.extend(
                test_variant(
                    '1.2. Include pattern', patterns=['!a/'],
                    expected=[
                        'a', 'x', 'a/b', 'a/b/c', 'x/a', 'x/x', 'x/a/b', 'x/x/a', 'x/x/x', 'x/x/a/b', 'x/x/x/a',
                        'x/x/x/a/b', 'x/x/x/a/b/c']))
            errors.extend(
                test_variant(
                    '1.3. Exclude, include pattern', patterns=['a', '!a'],
                    expected=[
                        'a', 'x', 'a/b', 'a/b/c', 'x/a', 'x/x', 'x/a/b', 'x/x/a', 'x/x/x', 'x/x/a/b', 'x/x/x/a',
                        'x/x/x/a/b', 'x/x/x/a/b/c']))
            errors.extend(
                test_variant(
                    '1.4. Include, exclude pattern', patterns=['!a', 'a'],
                    expected=[
                        'a', 'x', 'a/b', 'a/b/c', 'x/a', 'x/x', 'x/a/b', 'x/x/a', 'x/x/x', 'x/x/a/b', 'x/x/x/a',
                        'x/x/x/a/b', 'x/x/x/a/b/c']))

            errors.extend(
                test_variant(
                    '2.1. Exclude pattern', patterns=['**/a'],
                    expected=['x', 'a/b', 'a/b/c', 'x/x', 'x/a/b', 'x/x/x', 'x/x/a/b', 'x/x/x/a/b', 'x/x/x/a/b/c']))
            errors.extend(
                test_variant(
                    '2.2. Include pattern', patterns=['!**/a'],
                    expected=[
                        'a', 'x', 'a/b', 'a/b/c', 'x/a', 'x/x', 'x/a/b', 'x/x/a', 'x/x/x', 'x/x/a/b', 'x/x/x/a',
                        'x/x/x/a/b', 'x/x/x/a/b/c']))
            errors.extend(
                test_variant(
                    '2.3. Exclude, include pattern', patterns=['**/a', '!**/a'],
                    expected=[
                        'a', 'x', 'a/b', 'a/b/c', 'x/a', 'x/x', 'x/a/b', 'x/x/a', 'x/x/x', 'x/x/a/b', 'x/x/x/a',
                        'x/x/x/a/b', 'x/x/x/a/b/c']))
            errors.extend(
                test_variant(
                    '2.4. Include, exclude pattern', patterns=['!**/a', '**/a'],
                    expected=[
                        'a', 'x', 'a/b', 'a/b/c', 'x/a', 'x/x', 'x/a/b', 'x/x/a', 'x/x/x', 'x/x/a/b', 'x/x/x/a',
                        'x/x/x/a/b', 'x/x/x/a/b/c']))

            errors.extend(
                test_variant(
                    '3.1. Exclude pattern', patterns=['**/a/b/c'],
                    expected=[
                        'a', 'x', 'a/b', 'x/a', 'x/x', 'x/a/b', 'x/x/a', 'x/x/x', 'x/x/a/b', 'x/x/x/a', 'x/x/x/a/b']))
            errors.extend(
                test_variant(
                    '3.2. Include pattern', patterns=['!**/a/b/c'], 
                    expected=[
                        'a', 'x', 'a/b', 'a/b/c', 'x/a', 'x/x', 'x/a/b', 'x/x/a', 'x/x/x', 'x/x/a/b', 'x/x/x/a',
                        'x/x/x/a/b', 'x/x/x/a/b/c']))
            errors.extend(
                test_variant(
                    '3.3. Exclude, include pattern', patterns=['**/a/b/c', '!**/a/b/c'], 
                    expected=[
                        'a', 'x', 'a/b', 'a/b/c', 'x/a', 'x/x', 'x/a/b', 'x/x/a', 'x/x/x', 'x/x/a/b', 'x/x/x/a',
                        'x/x/x/a/b', 'x/x/x/a/b/c']))
            errors.extend(
                test_variant(
                    '3.4. Include, exclude pattern', patterns=['!**/a/b/c', '**/a/b/c'], 
                    expected=[
                        'a', 'x', 'a/b', 'a/b/c', 'x/a', 'x/x', 'x/a/b', 'x/x/a', 'x/x/x', 'x/x/a/b', 'x/x/x/a',
                        'x/x/x/a/b', 'x/x/x/a/b/c']))
        finally:
            chdir(working_directory_path)

        self.assertEqual(len(errors), 0, '\n\n' + '\n\n'.join(errors))

    def test_unanchored_file(self):
        """ Tests three forms of the unanchored file pattern

        These are the variants tested:

            `a`           includes directory or filename `a` in any sub-directory
            `**/a`        includes directory or filename `a` in any sub-directory
            `**/a/b/c`    include directory `a` with sub-directory `b` and file or directory `c` in any sub-directory

        This test runs against the contents of a fictional source directory.

        """
        filenames = OrderedDict([('a', 'f'), ('x/a', 'f'), ('x/x/a', 'f'), ('x/x/x/a/b/c', 'f')])
        test_variant, working_directory_path = self._prepare_source_directory(filenames)
        filenames = filenames.keys()

        try:
            errors = []

            # Variant 1

            errors.extend(
                test_variant(
                    '1.1. Exclude pattern', patterns=['a'],
                    expected=['x', 'x/x', 'x/x/x']))
            errors.extend(
                test_variant(
                    '1.2. Exclude, include pattern', patterns=['a', '!a'],
                    expected=['x', 'a', 'x/x', 'x/a', 'x/x/x', 'x/x/a', 'x/x/x/a', 'x/x/x/a/b', 'x/x/x/a/b/c']))
            errors.extend(
                test_variant(
                    '1.3. Exclude, include pattern', patterns=['a', '!*'],
                    expected=['x', 'a', 'x/x', 'x/a', 'x/x/x', 'x/x/a', 'x/x/x/a', 'x/x/x/a/b', 'x/x/x/a/b/c']))
            errors.extend(
                test_variant(
                    '1.4. Exclude, include pattern', patterns=['*', '!a'],
                    expected=['a', 'x/a', 'x/x/a', 'x/x/x/a', 'x/x/x/a/b', 'x/x/x/a/b/c']))
            errors.extend(
                test_variant(
                    '1.5. Include pattern', patterns=['!a'],
                    expected=['x', 'a', 'x/x', 'x/a', 'x/x/x', 'x/x/a', 'x/x/x/a', 'x/x/x/a/b', 'x/x/x/a/b/c']))
            errors.extend(
                test_variant(
                    '1.6. Include, exclude pattern', patterns=['!a', 'a'],
                    expected=['x', 'a', 'x/x', 'x/a', 'x/x/x', 'x/x/a', 'x/x/x/a', 'x/x/x/a/b', 'x/x/x/a/b/c']))
            errors.extend(
                test_variant(
                    '1.7. Include, exclude pattern', patterns=['!a', '*'],
                    expected=['a', 'x/a', 'x/x/a', 'x/x/x/a', 'x/x/x/a/b', 'x/x/x/a/b/c']))
            errors.extend(
                test_variant(
                    '1.8. Include, exclude pattern', patterns=['!*', 'a'],
                    expected=['x', 'a', 'x/x', 'x/a', 'x/x/x', 'x/x/a', 'x/x/x/a', 'x/x/x/a/b', 'x/x/x/a/b/c']))

            # Variant 2

            errors.extend(
                test_variant(
                    '2.1. Exclude pattern', patterns=['**/a'],
                    expected=['x', 'x/x', 'x/x/x', 'x/x/x/a/b', 'x/x/x/a/b/c']))
            errors.extend(
                test_variant(
                    '2.2. Include pattern', patterns=['!**/a'],
                    expected=['x', 'a', 'x/x', 'x/a', 'x/x/x', 'x/x/a', 'x/x/x/a', 'x/x/x/a/b', 'x/x/x/a/b/c']))
            errors.extend(
                test_variant(
                    '2.3. Exclude, include pattern', patterns=['**/a', '!**/a'],
                    expected=['x', 'a', 'x/x', 'x/a', 'x/x/x', 'x/x/a', 'x/x/x/a', 'x/x/x/a/b', 'x/x/x/a/b/c']))
            errors.extend(
                test_variant(
                    '2.4. Include, exclude pattern', patterns=['!**/a', '**/a'],
                    expected=['x', 'a', 'x/x', 'x/a', 'x/x/x', 'x/x/a', 'x/x/x/a', 'x/x/x/a/b', 'x/x/x/a/b/c']))

            # Variant 3

            errors.extend(
                test_variant(
                    '3.1. Exclude pattern', patterns=['**/a/b/c'], 
                    expected=['x', 'a', 'x/x', 'x/a', 'x/x/x', 'x/x/a', 'x/x/x/a', 'x/x/x/a/b']))
            errors.extend(
                test_variant(
                    '3.2. Include pattern', patterns=['!**/a/b/c'], 
                    expected=['x', 'a', 'x/x', 'x/a', 'x/x/x', 'x/x/a', 'x/x/x/a', 'x/x/x/a/b', 'x/x/x/a/b/c']))
            errors.extend(
                test_variant(
                    '3.3. Exclude, include pattern', patterns=['**/a/b/c', '!**/a/b/c'], 
                    expected=['x', 'a', 'x/x', 'x/a', 'x/x/x', 'x/x/a', 'x/x/x/a', 'x/x/x/a/b', 'x/x/x/a/b/c']))
            errors.extend(
                test_variant(
                    '3.4. Include, exclude pattern', patterns=['!**/a/b/c', '**/a/b/c'], 
                    expected=['x', 'a', 'x/x', 'x/a', 'x/x/x', 'x/x/a', 'x/x/x/a', 'x/x/x/a/b', 'x/x/x/a/b/c']))

            self.assertEqual(len(errors), 0, '\n\n' + '\n\n'.join(errors))

        finally:
            chdir(working_directory_path)

    def test_recursive_glob(self):
        """ Tests anchored and unanchored variants of the recursive glob file pattern

        These are the variants tested:
        .. code-block::
            l/**/m/n/*.foo
            **/o/**/p/*.bar

        This test runs against the contents of a fictional source directory.

        """
        pass

    def test_default_system_ignore_file(self):
        """ Ensures that the default ignore file works correctly

        This test runs against the contents of `$SPLUNK_HOME/lib/python2.7/site-packages/slim/config/ignore`. We setup
        a fake-home directory with no ~/.config/slim/ignore file to accomplish this.

        """
        pass

    def test_system_and_user_files_are_loaded_correctly(self):
        """ Ensures that the system and user files are correctly loaded

        This test runs against the contents of `$SPLUNK_HOME/lib/python2.7/site-packages/slim/config/ignore` and
        a manufactured version of ~/.config/slim/ignore in a fake-home directory to accomplish this.

        """
        pass

    # endregion

    # region Setup and teardown

    @classmethod
    def setUpClass(cls):

        cls._test_directory = mkdtemp(prefix='slim-test_ignore-', dir=environ['SPLUNK_SOURCE'])

        cls._fake_slim_home_directory = path.join(cls._test_directory, 'slim')
        cls._fake_slim_config_directory = path.join(cls._test_directory, 'slim', 'config')
        cls._fake_user_home_directory = path.join(cls._test_directory, 'home')
        cls._fake_user_config_directory = path.join(cls._fake_user_home_directory, '.config', 'slim')
        cls._fake_user_source_directory = path.join(cls._fake_user_home_directory, 'source')

        makedirs(cls._fake_slim_config_directory)
        makedirs(cls._fake_user_config_directory)
        makedirs(cls._fake_user_source_directory)

    @classmethod
    def tearDownClass(cls):
        rmtree(cls._test_directory)

    def setUp(self):
        environ[str('HOME')] = str(self._fake_user_home_directory)
        source = self._fake_user_source_directory
        if path.isdir(source):
            rmtree(source)
        makedirs(source)
        SlimConfigurationManager._initialize(self._fake_slim_home_directory)
        return

    def tearDown(self):
        environ[str('HOME')] = str(self._real_user_home_directory)
        source = self._fake_user_source_directory
        if path.isdir(source):
            rmtree(self._fake_user_source_directory)
        SlimConfigurationManager._initialize()

    # endregion

    # region Assert methods

    def assertIncludes(self, expected, source_directory, filenames):
        ignore = SlimIgnore(source_directory, source_directory)
        observed = sorted(list(ignore.ifilter(filenames)))
        expected = sorted([path.normcase(filename) for filename in sorted(expected)])
        self.assertListEqual(
            observed, expected, msg='Expected filenames ' + string(filenames) + ' to produce:\n    ' +
            string(expected) + '\n  not\n    ' + string(observed)
        )

    # endregion

    # region Protected

    _test_directory = None

    _fake_slim_home_directory = None
    _fake_slim_config_directory = None
    _fake_user_home_directory = None
    _fake_user_config_directory = None
    _fake_user_source_directory = None

    _real_slim_home_directory = path.join(string(environ['PYTHONPATH']), 'slim')
    _real_user_home_directory = path.expanduser('~')  # works across platforms

    @staticmethod
    def _copy(patterns, ignore_filename):
        with io.open(ignore_filename, encoding='utf-8', mode='w', newline='') as ostream:
            print('\n'.join(patterns), file=ostream)

    def _create_source_files(self, filenames):
        test_method_name = string(self._testMethodName)
        for filename, file_type in filenames.items():
            directory_path, name = path.split(filename)
            if directory_path:
                makedirs(directory_path)
            if name:
                if file_type == 'f':
                    with io.open(filename, mode='w', encoding='utf-8', newline='') as ostream:
                        print(test_method_name, file=ostream)
                elif file_type == 'd':
                    mkdir(filename)

    def _prepare_source_directory(self, filenames):

        source = self._fake_user_source_directory
        ignore_filename = path.join(source, '.slimignore')

        def test_variant(name, patterns, expected):
            self._copy(patterns, ignore_filename)
            errors = []
            try:
                self.assertIncludes(expected, source, filenames)
            except AssertionError as error:
                errors.append('Variant ' + name + ': ' + string(patterns) + ': ' + string(error))
            return errors

        working_directory = getcwd()
        chdir(source)

        try:
            self._create_source_files(filenames)
            start = len(source) + 1
            filenames = []
            for directory_path, directories, files in walk(source):
                if len(directory_path) < start:
                    filenames.extend(directories)
                    filenames.extend(files)
                else:
                    directory_path = directory_path[start:]
                    if len(directories) > 0:
                        filenames.extend(path.join(directory_path, name) for name in directories)
                    if len(files) > 0:
                        filenames.extend(path.join(directory_path, name) for name in files)
        except:
            chdir(working_directory)
            raise

        return test_variant, working_directory

    # endregion
    pass


if __name__ == '__main__':
    import unittest.main
    unittest.main()
