#!/usr/bin/env python
# coding=utf-8
#
# Copyright © 2015 Splunk, Inc. All Rights Reserved.

from __future__ import absolute_import, division, print_function, unicode_literals

from unittest import TestCase
from itertools import permutations
from string import ascii_letters, digits

import importlib
import io
import json
import os
import shutil
import stat
import sys
import tarfile
import tempfile

from test_slim import generate_random_string

__all__ = ['TestArgumentParser']


class TestArgumentParser(TestCase):

    # region Special methods

    def __init__(self, method_name):
        TestCase.__init__(self, method_name)
        self._test_name = '.'.join((
            os.path.splitext(os.path.basename(__file__))[0], type(self).__name__, self._testMethodName))

    # endregion

    # region Setup/teardown methods

    @classmethod
    def setUpClass(cls):
        cls.command = importlib.import_module('partition')
        cls.command.parser.prog = 'partition'
        cls.options = {tuple(action.option_strings): action.required for action in cls.command.parser._actions}

    def tearDown(self):
        self._enable_all_options()

    # endregion

    # region assert methods

    def assertSystemExit(self, code, function, *args):

        self.assertTrue(callable(function))

        call = function.__name__ + '(' + ' '.join(*args) + ')'
        fake_stderr = io.StringIO()
        real_stderr = sys.stderr
        sys.stderr = fake_stderr

        try:
            function(*args)
        except SystemExit as error:
            self.assertEqual(code, error.code, 'Expected exit code {0}, not {1}: {2}'.format(code, error.code, call))
        except BaseException as error:
            self.fail('Expected SystemExit, not {0}: {1}'.format(type(error).__name__, call))
        else:
            self.fail('Expected SystemExit, not success: {0}'.format(call))
        finally:
            sys.stderr = real_stderr

    # endregion

    # region Test methods

    def test_option_debug(self):

        option_strings = ('--debug',)
        self._disable_all_options_except(option_strings)

        parse_args = self.command.parser.parse_args

        for option_name in option_strings:
            arguments = parse_args([option_name])
            self.assertIs(True, arguments.debug)

        arguments = parse_args([])
        self.assertIs(False, arguments.debug)

    def test_option_help(self):

        self._enable_all_options()

        with io.StringIO(newline=None) as fake_stdout:
            real_stdout = sys.stdout
            sys.stdout = fake_stdout
            try:
                self.command.parser.parse_args(['--help'])
            except SystemExit as system_exit:
                text = fake_stdout.getvalue()
                self.assertEqual(system_exit.code, 0)
                self.assertIs(text.startswith('usage: partition [-h] [-c] [--debug]'), True)
            except BaseException as exception:
                self.fail('Expected SystemExit(0), not {0}({1})'.format(type(exception).__name__, exception))
            else:
                self.fail('Expected SystemExit(0), not no exception')
            finally:
                sys.stdout = real_stdout

    # noinspection PyBroadException
    def test_option_deployment_package_fail(self):

        option_strings = ('-d', '--deployment')
        self._disable_all_options_except(option_strings)

        parse_args = self.command.parser.parse_args

        for option_name in option_strings:

            # Poorly formed json

            self.assertSystemExit(1, parse_args, [option_name, '{"name": "foo", "workload": "forwarder", "inputGroups": ["*"]'])

            # Illegal field names

            self.assertSystemExit(1, parse_args, [option_name, '{"not_name": "illegal", "name": "foo", "workload": "forwarder", "inputGroups": ["*"]}'])

            # Illegal name value

            specification = '{{"name": {0}, "workload": "forwarder", "inputGroups": ["*"]}}'

            for value in ('"*foo"', '"f*oo"', '"foo*"', '{}', '[]', '1', 'true', 'false', 'null'):
                self.assertSystemExit(1, parse_args, [option_name, specification.format(value)])

            # Illegal workload value

            specification = '{{"name": "foo", "workload": {0}, "inputGroups": ["*"]}}'

            for value in ('"undefined"', '[]', '["undefined"]', '["undefined", "forwarder"]', '{}', '1', 'true', 'false', 'null'):
                self.assertSystemExit(1, parse_args, [option_name, specification.format(value)])

            # Illegal forwarderGroups value

            specification = '{{"name": "foo", "workload": {0}, "inputGroups": {1}}}'

            for value in ('{}', '1', 'true', 'false', 'null'):
                self.assertSystemExit(1, parse_args, [option_name, specification.format('"forwarder"', value)])

            self.assertSystemExit(1, parse_args, [option_name, specification.format('["searchHead", "indexer"]', '["*"]')])

    def test_option_deployment_package_pass(self):

        option_strings = ('-d', '--deployment')
        self._disable_all_options_except(option_strings)

        parse_args = self.command.parser.parse_args

        safe_filename = ascii_letters + digits + ' _.-'

        for option_name in option_strings:

            # Legal name values

            specification = '{{"name": "{0}", "workload": "forwarder", "inputGroups": ["*"]}}'

            for value in safe_filename:
                ns = parse_args([option_name, specification.format(value)])
                self.assertEqual(value, ns.deployment_packages[0].name)

            ns = parse_args([option_name, specification.format(safe_filename)])

            # Legal workload values

            specification = '{{"name": "foo", "workload": "{0}"}}'

            for value in ('searchHead', 'indexer', 'forwarder'):
                ns = parse_args([option_name, specification.format(value)])
                self.assertEqual(frozenset((value,)), ns.deployment_packages[0].workload)

            specification = '{{"name": "foo", "workload": {0}}}'

            for count in range(1, 4):
                for value in permutations(["searchHead", "indexer", "forwarder"], count):
                    ns = parse_args([option_name, specification.format(json.dumps(value))])
                    self.assertEqual(frozenset(value), ns.deployment_packages[0].workload)

            # Legal forwarderGroups values

            specification = '{{"name": "foo", "workload": {0}, "inputGroups": {1}}}'

            for workload_count in range(1, 3):

                for workload in permutations(['searchHead', 'indexer'], workload_count):
                    workload += ('forwarder', )  # because the forwarder workload is required for forwarderGroups

                    for forwarder_group_count in range(1, 4):
                        input_groups = [generate_random_string(8, 64) for _ in range(0, forwarder_group_count)]
                        ns = parse_args(
                            [option_name, specification.format(json.dumps(workload), json.dumps(input_groups))])
                        self.assertEqual(frozenset(input_groups), ns.deployment_packages[0].inputGroups)

                    input_groups = generate_random_string(8, 64)
                    ns = parse_args([option_name, specification.format(json.dumps(workload), json.dumps(input_groups))])
                    self.assertEqual(frozenset([input_groups]), ns.deployment_packages[0].inputGroups)

            input_groups = generate_random_string(8, 64)
            ns = parse_args([option_name, specification.format(json.dumps('forwarder'), json.dumps(input_groups))])
            self.assertEqual(frozenset([input_groups]), ns.deployment_packages[0].inputGroups)

            ns = parse_args([option_name, specification.format(json.dumps('forwarder'), json.dumps('*'))])
            self.assertEqual(frozenset(('*',)), ns.deployment_packages[0].inputGroups)

            ns = parse_args([option_name, specification.format(json.dumps('forwarder'), json.dumps(['*']))])
            self.assertEqual(frozenset(('*',)), ns.deployment_packages[0].inputGroups)

            ns = parse_args([option_name, specification.format(json.dumps('forwarder'), json.dumps(['*', 'foo']))])
            self.assertEqual(frozenset(('*',)), ns.deployment_packages[0].inputGroups)

    def test_output_dir_fail(self):

        option_strings = ('-o', '--output-dir')

        root_dir = tempfile.mkdtemp(prefix=self._test_name)
        self._disable_all_options_except(option_strings)

        try:
            output_dir = os.path.join(root_dir, 'output-dir')
            parse_args = self.command.parser.parse_args

            for option_string in option_strings:

                # Can't create directory because permission is denied

                if sys.platform == 'win32':
                    pass  # TODO: Use win32security module to set DACLs on the directory because os.chmod is ineffective
                else:
                    os.chmod(root_dir, stat.S_IREAD)
                    self.assertSystemExit(1, parse_args, [option_string, output_dir])

                # Can't create directory because a file (not a directory) by that name already exists

                os.chmod(root_dir, stat.S_IREAD | stat.S_IWRITE | stat.S_IEXEC)

                with io.open(output_dir, encoding='utf-8', mode='w', newline='') as f:
                    f.write(self._test_name)

                self.assertSystemExit(1, parse_args, [option_string, output_dir])
        finally:
            shutil.rmtree(root_dir, ignore_errors=True)

    def test_output_dir_pass(self):

        option_strings = ('-o', '--output-dir')
        self._disable_all_options_except(option_strings)
        root_dir = tempfile.mkdtemp(prefix=self._test_name)

        try:
            output_dir = os.path.join(root_dir, 'parent', 'output-dir')
            parse_args = self.command.parser.parse_args

            for option_string in option_strings:

                # Can create a directory, making parent directories as needed

                ns = parse_args([option_string, output_dir])
                self.assertEqual(output_dir, ns.output_dir)

                # Can access an existing directory

                ns = parse_args([option_string, output_dir])
                self.assertEqual(output_dir, ns.output_dir)

        finally:
            shutil.rmtree(root_dir)

    # endregion

    # region Helper methods

    def _disable_all_options_except(self, option_strings):

        actions = TestArgumentParser.command.parser._actions
        under_test = list(option_strings)

        for action in actions:
            if action.option_strings == under_test:
                if action.option_strings == ['--positional-argument']:
                    action.option_strings = []
                action.required = TestArgumentParser.options[option_strings]
            else:
                if len(action.option_strings) == 0:
                    action.option_strings = ['--positional-argument']
                action.required = False

    def _enable_all_options(self):
        for action in TestArgumentParser.command.parser._actions:
            if action.option_strings == ['--positional-argument']:
                action.option_strings = []
            action.required = TestArgumentParser.options[tuple(action.option_strings)]

    command = None
    options = None

    # endregion
    pass
