#!/usr/bin/env python
# coding=utf-8
#
# Copyright © Splunk, Inc. All Rights Reserved.

""" Command-specific arguments

    {
        name: uninstall
        args: {
            app_id: <app_id>
        }
    },
    {
        name: install
        args: {
            app_package: <app_package>
            combine_search_head_indexer_workloads: <0|1>
            workloads: {
                <forwarder_group_X>: [<server_class_A>, <server_class_B>, ...]
                <forwarder_group_Y>: [<server_class_C>, ...]
                ...
            }
        }
    },
    {
        name: update
        args: {
            app_package: <app_package>
        }
    }

"""


from __future__ import absolute_import, division, print_function, unicode_literals

from os import path
import sys
import json

from slim.utils import *
from slim.app import *
from slim.command import *

# Argument parser definition

parser = SlimArgumentParser(description='''Update an installation graph given the actions defined.''')

parser.add_repository()
parser.add_installation()
parser.add_output_directory(description='installation graph and deployment packages')

parser.add_argument(
    '-a', '--actions',
    type=SlimInstallationActionArgument(), nargs='+', default=[],
    help='list of actions to be applied to the given installation graph: install, update, uninstall',
    metavar='<action> [<action>...]')

parser.add_argument(
    '--disable-automatic-resolution',
    action='store_const', const=True, default=False,
    help='do not automatically resolve dependency conflicts by updating installed versions')


def main(argv):

    args = parser.parse_args(argv[1:])
    server_collection = _update_installation(
        args.installation, args.repository, args.actions, args.disable_automatic_resolution
    )

    filename = path.join(args.output_dir, 'installation-update.json')
    server_collection.save(filename)
    SlimLogger.exit_on_error()

    SlimLogger.information('Saved updated installation graph to ', encode_filename(filename))


def update_installation(installation_graph, action_list, disable_automatic_resolution=False):

    stream_type = SlimStringIOArgument(name="installation_graph.json")

    with stream_type(value=installation_graph) as istream:
        action_type = SlimInstallationActionArgument()
        actions = []
        for action in action_list:
            actions.append(action_type(value=json.dumps(action)))

        server_collection = _update_installation(
            istream, slim_configuration.repository_path, actions, disable_automatic_resolution
        )

    # Save the installation graph to the payload only (not file output)
    server_collection.save(None)
    SlimLogger.exit_on_error()


# pylint: disable=redefined-builtin
def _update_installation(file, repository, actions, disable_automatic_resolution):

    SlimLogger.step('Updating installation graph at ', encode_filename(file.name))

    server_collection = AppServerClassCollection.load(file, repository)
    SlimLogger.exit_on_error()

    for action in actions:
        server_collection.update_installation(action, disable_automatic_resolution)
        SlimLogger.exit_on_error()

    return server_collection


if __name__ == '__main__':
    # noinspection PyBroadException
    try:
        main(sys.argv)
    except SystemExit:
        raise
    except:
        SlimLogger.fatal(exception_info=sys.exc_info())
